/**
 * We have implemented a new location-based compare engine, which may work
 * better than classic compare engine, but requires much more resources snd thus
 * cannot be used in history algorithm. However, in can be used as a stand-alone
 * tool, and this script implements rendering the results of this script
 */

/**
 * this variable contains links to all elements displaying stats percentage
 * along with their stats.
 * It is used to rebuild such elements when user changes display settings 
 */
var AC_ALL_STATS_CONTAINERS=[];

function fetchCompareData(){
	$('#warningsLoading').show();
	$('#warningsLoadingCounter').text("");
	setTimeout(fetchCompareDataImpl, 1);
}

function fetchCompareDataImpl(){
 	WarningObjectsWrap.WarningObjects={};
 	var WarningObjects = WarningObjectsWrap.WarningObjects;
	try {
		var starttime = new Date().getTime();
		if (!fetchTypesValues(false, COMPARE_PROJECTS.project_1)){
			fetching = false;
			return;
		}
		if (!fetchTypesValues(false, COMPARE_PROJECTS.project_2)){
			fetching = false;
			return;
		}
		var typetime = new Date().getTime();
		var Checkers = null;
		OK = false;
		var url='/history/api?act=get_warnings_compare'+
			getProjectParameter(COMPARE_PROJECTS.project_1, '_1')+
			getBranchParameter(COMPARE_PROJECTS.project_1, '_1')+
			'&snapshot_id_1='+makeSnapshotIDParm(COMPARE_PROJECTS.project_1)+
			getProjectParameter(COMPARE_PROJECTS.project_2, '_2')+
			getBranchParameter(COMPARE_PROJECTS.project_2, '_2')+
			'&snapshot_id_2='+makeSnapshotIDParm(COMPARE_PROJECTS.project_2)+
			'&is_group_similar='+GROUP_SIMILAR;
		
		$.ajax( {
			url: url,
			data: {config:$("#compconfig").val()},
			type: 'POST',
			dataType: 'json',
			cache: false,
			async: false,
			success: function (data, textStatus) {
				//data is already json
				if (data.status === "OK") {
					OK = true;
					var endtime = new Date().getTime();
					console.log('Loading time: ' + (endtime-starttime)+"ms total, "+(typetime-starttime)+"ms types");
					Checkers = data.result;
					if (Checkers!=null) {
						for (var idx=0; idx<Checkers.length; idx++){
							var Checker = Checkers[idx];
							var checker_warnings={missing:0, both_pessimistic:0, both_semi:0, both_optimistic:0, extra:0};
							if (Checker.missing!=undefined){
								for (var widx=0; widx<Checker.missing.length; widx++){
									var Warning = Checker.missing[widx];
									makeWarningSortingData(Warning);
									Warning.project = COMPARE_PROJECTS.project_1;
									rememberWarning(Warning, WarningObjects)
									checker_warnings.missing++;
								}
							}
							var process = function(both, isPrime){
								for (var cidx=0; cidx<both.length; cidx++){
									var subChecker = both[cidx].data;
									for (var widx=0; widx<subChecker.length; widx++){
										var Warning = subChecker[widx].first;
										makeWarningSortingData(Warning);
										Warning.project = COMPARE_PROJECTS.project_1;
										rememberWarning(Warning, WarningObjects)

										Warning = subChecker[widx].second;
										makeWarningSortingData(Warning);
										Warning.project = COMPARE_PROJECTS.project_2;
										rememberWarning(Warning, WarningObjects)
										if (isPrime==true){
											checker_warnings.both_pessimistic++;
										} else if (isPrime==false) {
											checker_warnings.both_optimistic++;
										} else if (isPrime==null) {
										    //sorry, the only way to converts binary logic into ternary logic in the fast way.
											checker_warnings.both_semi++;
										}
									}
								}
							}
							if (Checker.both_optimistic!=undefined){
								process(Checker.both_optimistic, false);
							}
							if (Checker.both_pessimistic!=undefined){
								process(Checker.both_pessimistic, true);
							}
							if (Checker.both_semi!=undefined){
								process(Checker.both_semi, null);
							}
							if (Checker.extra!=undefined){
								for (var cidx=0; cidx<Checker.extra.length; cidx++){
									var subChecker = Checker.extra[cidx].data;
									for (var widx=0; widx<subChecker.length; widx++){
										var Warning = subChecker[widx];
										makeWarningSortingData(Warning);
										Warning.project = COMPARE_PROJECTS.project_2;
										rememberWarning(Warning, WarningObjects)
										checker_warnings.extra++;
									}
								}
							}
							Checker.checker_warnings=checker_warnings;
						}
						renderComparedCheckers(Checkers);
						gotWarnings = true;
					}
				}
			},
			error: ajaxError,
			complete:function(){
				fetching = false;
				}
		});
		$('#warningsLoadingCounter').text("");
	} finally {
		$('#warningsLoading').hide();
	}
}

function renderComparedCheckers(Checkers) {
	var chNum = Checkers.length;
	if (chNum==0) {
		$('#warningsList').text("No warnings");
	}
	//we may drop this now because it is broken while we remembered displayed warnings. Do not try at home.
	DisplayedWarningsCheckersData = {};
	var TotalStats={missing:0, both_pessimistic:0, both_semi:0, both_optimistic:0, extra:0};
	AC_ALL_STATS_CONTAINERS = [];
	for (var i = 0; i < chNum; i++) {
		var Checker = Checkers[i];

		getCheckerGroupBox(Checker.checker_id, Checker);

		var cw = Checker.checker_warnings;
		TotalStats.missing+=cw.missing;
		TotalStats.both_pessimistic+=cw.both_pessimistic;
		TotalStats.both_semi+=cw.both_semi;
		TotalStats.both_optimistic+=cw.both_optimistic;
		TotalStats.extra+=cw.extra;
	}
	resortCheckerGroups();
	$("#ac_stats").html("");
	$("#ac_stats").append(makeStatsPercentage(TotalStats));
	makePercentageColumnEqualWidth();
}

function makePercentageColumnEqualWidth(){
	//And now let's set all 'percentage' boxes equal width:
	var greatestWidth = 0;
	$(".ac_percentage").each(function() {
	    var theWidth = $(this).width();
	    if( theWidth > greatestWidth) {
	        greatestWidth = theWidth;
	    }
	});
	$(".ac_percentage").width(greatestWidth+5);
}

function makeStatsPercentage(stats, Checker){
	var result="";
	//total contains number of first-project warnings
	var total = stats.missing+stats.both_optimistic+stats.both_pessimistic+stats.both_semi;
	var percentage;
	if (total==0){
		percentage=100;
	} else {
		var found = 0;
		if (RECALL_COUNT_PESS){
			found += stats.both_pessimistic;
		}
		if (RECALL_COUNT_SEMI){
			found += stats.both_semi;
		}
		if (RECALL_COUNT_OPTI){
			found += stats.both_optimistic;
		}
		percentage=Math.round(found*100/total);
	}
	var redHexString=Math.round(255*Math.sqrt((100-percentage)/100)).toString(16);
	if (redHexString.length==1)
		redHexString='0'+redHexString;
	var grnHexString=Math.round(255*Math.sqrt(percentage/100)).toString(16);
	if (grnHexString.length==1)
		grnHexString='0'+grnHexString;
	result+="<div class='ac_percentage' style='display:inline-block;' title='-"+stats.missing+"|~~"+stats.both_optimistic+"|~="+stats.both_semi+"|=="+stats.both_pessimistic+"|+"+stats.extra+"'>";
	result+="<font color='#"+redHexString+grnHexString+"00'>(";
	if (total==0){
		result += "----"
	} else {
		result+=percentage+"%";
	}
	result+=")</font></div>";
	
	result = $(result);
	
	AC_ALL_STATS_CONTAINERS.push({element:result, stats: stats, Checker: Checker});
	
	if (Checker != undefined){
		//we make data for checker sort engine
		
		//amount contains number of second-project warnings
		var amount = stats.both_optimistic+stats.both_pessimistic+stats.both_semi+stats.extra;
		var missing = total-found;
		
		DisplayedWarningsCheckersData[Checker.checker_id]={
				_id: Checker.checker_id,
				name: getCheckerName(Checker.checker_id),
				amount: amount,
				miss_num: missing,
				miss_prc: ((total==0)?-1:(missing/total)),
				match_num: found,
				match_prc: ((total==0)?1.1:(found/total)),
				extra_num: stats.extra
		};
	}
	
	return result;
}

function getCheckerGroupBox(checkerID, data) {
	var id = "LIST_"+checkerID;
	var a = document.getElementById(id);
	if (a==null) {
		var containerID = id+"_container";
		var container = document.createElement('div');
		container.setAttribute('id', containerID);
		var list = document.getElementById('warningsList');
		var checkerName = getCheckerName(checkerID);
		insertInIDOrder(list, container, checkerName);
		
		var cw=data.checker_warnings;
		
		container.innerHTML = "<div id='"+id+"_header' class='GROUP_HEADER' title='"+checkerName+getCheckerDesc(checkerID)+"'>"+checkerName+
		/*"(<div id='ch_"+warnClassID+"_counter' class='counter' style='display:inline'>0</div>)"+*/
		" <div id='"+id+"_expansion_status' style='display:inline'>&gt;</div></div>"+
		                      "<div id='"+id+"' style='display:none;'></div>";
		$("#"+id+"_header").prepend(makeStatsPercentage(cw, data));
		
		var blockShown=false;

		// _header also affects displayWarningByIDAndClass
		$('#'+id+'_header').click(function() {
			blockShown=!blockShown;
			if (blockShown) {
				if ( $('#'+id).children().length == 0 ) {
					$('#warningsLoading').show();
					var progress_counter=0;
					var progress_total=cw.missing+cw.both_optimistic+cw.both_semi+cw.both_pessimistic+cw.extra;
					makelvl2entry=function(suffix, name){
						return "<div id='"+id+"_"+suffix+"_header' class='GROUP2_HEADER'>"+name+
						/*"(<div id='ch_"+warnClassID+"_counter' class='counter' style='display:inline'>0</div>)"+*/
						" <div id='"+id+"_"+suffix+"_expansion_status' style='display:inline'>&gt;</div></div>"+
						                      "<div id='"+id+"_"+suffix+"' style='display:none;'></div>";
					};
					var a = document.getElementById(id);
					if (data.both_optimistic!=undefined)
						a.innerHTML += makelvl2entry('both_optimistic', '('+cw.both_optimistic+')Both have (optimistic)');
					if (data.both_semi!=undefined)
						a.innerHTML += makelvl2entry('both_semi', '('+cw.both_semi+')Both have (semi-pessimistic)');
					if (data.both_pessimistic!=undefined)
						a.innerHTML += makelvl2entry('both_pessimistic', '('+cw.both_pessimistic+')Both have (pessimistic)');
					if (data.missing!=undefined)
						a.innerHTML += makelvl2entry('miss', '('+cw.missing+')Missing');
					if (data.extra!=undefined)
						a.innerHTML += makelvl2entry('extra', '('+cw.extra+')Extra');
					if (data.missing!=undefined){
						var classcollection = data.missing;
						classcollection = classcollection.sort(function(a,b){
							return a.ui_sortingdata.localeCompare(b.ui_sortingdata);
						});
						var idm = id+"_miss";
						if (classcollection.length>0){
							var container = document.getElementById(idm);
							for (var idx=0; idx<classcollection.length; idx++){
								makeWarningBox(classcollection[idx], container);
								if ((progress_counter++%977)==0) //no frequent updates==more happy browsers. constant is magic.
									$('#warningsLoadingCounter').text(" "+progress_counter+"/"+progress_total);
							}
						}
						var missShown=false;
						var header = $('#'+idm+'_header'); 
						header.click(function() {
							missShown=!missShown;
							if (missShown) {
								$('#'+idm).show();
								$('#'+idm+'_expansion_status').text("v");
							} else {
								$('#'+idm).hide();
								$('#'+idm+'_expansion_status').text(">");
							}
						});
						//right-click gives popup with list of warning id's in this group. This may help you to perform sampling: use these warning-id's in filter.
						addDisplayWIDsListener(header, data.missing);
					}
					
					render_both= function(checkercollection, idb){
						if (checkercollection.length>0){
							var bcontainer = document.getElementById(idb);

							for (var chidx=0; chidx<checkercollection.length>0; chidx++){
								var element = checkercollection[chidx];
								var checkerName = getCheckerName(element.checker_id);
								
								(function(){
								var idbc = idb+"_"+element.checker_id;
								var checkercontainer = document.createElement('div');
								checkercontainer.setAttribute('id', idbc+"_container");
								var list = document.getElementById('warningsList');

								var sortingdata = Array(8-element.data.length.toString().length).join("0")+element.data.length;

								insertInIDOrder(bcontainer, checkercontainer, sortingdata, true);
								
								checkercontainer.innerHTML = "<div id='"+idbc+"_header' class='GROUP3_HEADER'>"+
								"("+element.data.length+")"+checkerName+
								/*"(<div id='ch_"+warnClassID+"_counter' class='counter' style='display:inline'>0</div>)"+*/
								" <div id='"+idbc+"_expansion_status' style='display:inline'>&gt;</div></div>"+
								                      "<div id='"+idbc+"' style='display:none;'></div>";
								
								var classcollection = element.data;
								classcollection = classcollection.sort(function(a,b){
									return a.first.ui_sortingdata.localeCompare(b.first.ui_sortingdata);
								});
								if (classcollection.length>0){
									var container = document.getElementById(idbc);
									var odd=false;
									for (var idx=0; idx<classcollection.length; idx++){
										odd=!odd;
										warningGroup = document.createElement('div');
										warningGroup.setAttribute('class', "WARNING_CONTAINER");
										warningGroup.setAttribute('style', "background:"+(odd?"#cccccc":"#eeeeee"));
										container.appendChild(warningGroup);
										
										makeWarningBox(classcollection[idx].first, warningGroup);
										makeWarningBox(classcollection[idx].second, warningGroup);
										if ((progress_counter++%977)==0) //no frequent updates==more happy browsers. constant is magic.
											$('#warningsLoadingCounter').text(" "+progress_counter+"/"+progress_total);
									}
								}
								var bchShown=false;
								$('#'+idbc+'_header').click(function() {
									bchShown=!bchShown;
									if (bchShown) {
										$('#'+idbc).show();
										$('#'+idbc+'_expansion_status').text("v");
									} else {
										$('#'+idbc).hide();
										$('#'+idbc+'_expansion_status').text(">");
									}
								});
								})();
							}
						}
						var bothShown=false;
						$('#'+idb+'_header').click(function() {
							bothShown=!bothShown;
							if (bothShown) {
								$('#'+idb).show();
								$('#'+idb+'_expansion_status').text("v");
							} else {
								$('#'+idb).hide();
								$('#'+idb+'_expansion_status').text(">");
							}
						});						
					}
					if (data.both_pessimistic!=undefined){
						render_both(data.both_pessimistic, id+"_both_pessimistic");
					}
					if (data.both_semi!=undefined){
						render_both(data.both_semi, id+"_both_semi");
					}
					if (data.both_optimistic!=undefined){
						render_both(data.both_optimistic, id+"_both_optimistic");
					}
					if (data.extra!=undefined){
						var checkercollection = data.extra;
						var ide = id+"_extra";
						if (checkercollection.length>0){
							var econtainer = document.getElementById(ide);

							for (var chidx=0; chidx<checkercollection.length>0; chidx++){
								var element = checkercollection[chidx];
								var checkerName = getCheckerName(element.checker_id);
								
								(function(){
								var idec = ide+"_"+element.checker_id;
								var checkercontainer = document.createElement('div');
								checkercontainer.setAttribute('id', idec+"_container");
								var list = document.getElementById('warningsList');

								var sortingdata = Array(8-element.data.length.toString().length).join("0")+element.data.length;

								insertInIDOrder(econtainer, checkercontainer, sortingdata, true);
								
								checkercontainer.innerHTML = "<div id='"+idec+"_header' class='GROUP3_HEADER'>"+
								"("+element.data.length+")"+checkerName+
								/*"(<div id='ch_"+warnClassID+"_counter' class='counter' style='display:inline'>0</div>)"+*/
								" <div id='"+idec+"_expansion_status' style='display:inline'>&gt;</div></div>"+
								                      "<div id='"+idec+"' style='display:none;'></div>";
								
								var classcollection = element.data;
								classcollection = classcollection.sort(function(a,b){
									return a.ui_sortingdata.localeCompare(b.ui_sortingdata);
								});
								if (classcollection.length>0){
									var container = document.getElementById(idec);
									for (var idx=0; idx<classcollection.length; idx++){
										makeWarningBox(classcollection[idx], container);
										if ((progress_counter++%977)==0) //no frequent updates==more happy browsers. constant is magic.
											$('#warningsLoadingCounter').text(" "+progress_counter+"/"+progress_total);
									}
								}
								var echShown=false;
								var header = $('#'+idec+'_header'); 
								header.click(function() {
									echShown=!echShown;
									if (echShown) {
										$('#'+idec).show();
										$('#'+idec+'_expansion_status').text("v");
									} else {
										$('#'+idec).hide();
										$('#'+idec+'_expansion_status').text(">");
									}
								});
								//right-click gives popup with list of warning id's in this group. This may help you to perform sampling: use these warning-id's in filter.
								addDisplayWIDsListener(header, classcollection);
								})();
							}
						}
						var extraShown=false;
						$('#'+ide+'_header').click(function() {
							extraShown=!extraShown;
							if (extraShown) {
								$('#'+ide).show();
								$('#'+ide+'_expansion_status').text("v");
							} else {
								$('#'+ide).hide();
								$('#'+ide+'_expansion_status').text(">");
							}
						});
					}
					$('#warningsLoadingCounter').text("");
					$('#warningsLoading').hide();
				}
				$('#'+id).show();
				$('#'+id+'_expansion_status').text("v");
			} else {
				$('#'+id).hide();
				$('#'+id+'_expansion_status').text(">");
			}
		})
		res = document.getElementById(id);
		return res;
	}
	return a;
}

function acRecalculateStats(){
	var oldContainers = AC_ALL_STATS_CONTAINERS;
	AC_ALL_STATS_CONTAINERS = [];
	DisplayedWarningsCheckersData = {};
	for (var i = 0; i < oldContainers.length; i++){
		var data = oldContainers[i];
		data.element.replaceWith(makeStatsPercentage(data.stats, data.Checker));
	}
	
	makePercentageColumnEqualWidth();

	resortCheckerGroups();
}