/**
 * TODO: check that ll callers do know the fact all thic creators do create
 * incomplete objects and do fetch their data asynchronously..
 */

/**
 * This function initializes object responsible for project's snapshot list.
 * If a second argument (field) is provided, that jQuery component will be filled with snapshot selector. 
 * @param project
 * @param field
 * @param branch may be null, then default branch value will be used
 * @returns
 */
function makeSnapshotController(project, field, branch, selectionChangeListener){
	var list = null;
	if (field != null){
		list = $("<select size=1 style='width:100'/>");
		field.html("");
		field.append(list);
		if (project == null)
			list.append($("<option>----</option>"));
		else
			list.append($("<option>Loading...</option>"));
		list.prop('disabled', true);
	}
	//in case if we just had to clean up data...
	if (project == null)
		return null;
	
	var branchBuilds = null;
	/**
	 * If snapshots were received. Even if an empty list was received.
	 */
	var controller = {
		gotBulds: false,
		firstPaint: true,
		fetch: function(){
			if (branchBuilds!=null&&(branchBuilds.last.data.previous_snapshot_id==null)){
				return;
			}
			controller.fetchBuilds(branchBuilds, function(newBuildList) {
				branchBuilds=newBuildList;
				controller.repaintList();
			}, ajaxError);
		},
		
		repaintList: function(){
			//if no UI element was provided, we have nothing to repaint
			if (list==null){
				return;
			}
			list.unbind();
			list.empty();
			var currentBuild = branchBuilds;
			if (currentBuild==null){
				var html="<option value=''>No snapshots</option>";
				list.append($(html));
				list.prop('disabled', true);
			} else {
				list.prop('disabled', false);
				var hasMoreData = true;
				while (currentBuild!=null){
					var html="<option value=\""+currentBuild.data.snapshot_id+"\">"+currentBuild.data.description+"</option>";
					list.append($(html));
					if ((currentBuild.next==undefined)&&(currentBuild.data.previous_snapshot_id==null))
						hasMoreData = false;
					currentBuild = currentBuild.next;
				}
				if (hasMoreData){
					list.append($("<option value=\"...\">Fetch more...</option>"));
				}
				list.change(function() {
					controller.switchBuild();
				});
				if (controller.firstPaint){
					list.change();
				}
				controller.firstPaint = false;
			}
		},
		
		switchBuild: function(){
			if (list==null){
				//Cannot happen; this should only be called from UI change listener
				return;
			}
			//check if 'Fetch more' is selected...
			var data = list.val();
			if (data=='...'){
				controller.fetch(project);
				return;
			}

			if (selectionChangeListener!=null){
				//I can has async!
				setTimeout(function(){
					selectionChangeListener(controller.getSelectedSnapshotID())
				}, 1);
			}
		},

		// Update a linked list of builds on a given branch of a given project.
		// The list is updated by appending to the end, with as many snapshots
		// as the server reports for one request.
		// Argument onError is a function with the same prototype as ajaxError
		// that handles an ajax error.
		fetchBuilds: function (buildList, useExtendedList, onError){
			function generateBranchParameter(){
				try {
					if (branch != null){
						project.ui.branch_override = branch;
					}
					return getBranchParameter(project);
				} finally {
					delete project.ui.branch_override;
				}
			}
			$.ajax( {
				url:'/history/api?act=get_snapshots'
					+getProjectParameter(project)
					+(buildList==null?'':'&from='+buildList.last.data.previous_snapshot_id)
					+generateBranchParameter(),
				dataType: 'json',
				type: 'GET',
				cache: false,
				success: function (data, textStatus) {
					//data is already json
					if (data.status != "OK") {
						setError("Error:"+data.error);
					} else {
						var builds = data.snapshots;
						for (var i = 0; i < builds.length; i++) {
							build = builds[i];
							if (buildList==null) {
								if (useExtendedList==null){
									//we create new data structure, but the caller won't get it. Inform caller?
								}
								buildList = new Object();
								buildList.last = buildList;
								buildList.numBuilds = 1;
							} else {
								buildList.last.next = new Object();
								buildList.last = buildList.last.next; 
								buildList.numBuilds += 1;
							}
							buildList.last.data = build;
						}
						useExtendedList(buildList);
					}
				},
				error: onError,
				complete: function(){
					controller.gotBuilds = true;
				}
			});
		},
		
		getSelectedSnapshotID: function(){
			if (list==null){
				//no UI: we return latest snapshot
				if (!controller.gotBuilds){
					//TODO: no data was loaded (yet?). Time to load it yourself.
					//Actually in current setup should not havven since 'init' fetches everything and is a synchronous method
					setError("Some async problem...");
					return null;
				}
				if (branchBuilds==null){
					//no data...
					return null;
				}
				return branchBuilds.data.snapshot_id;
			} else {
				var requestedBuild = list.val();

				if (requestedBuild=='...') //special case, but usually should not happen
					requestedBuild = '';
				
				//TODO: check if throwing exception on '' won't break anything
				if (requestedBuild==null) {
					throw "No snapshot list available (No analysis data in selected branch?)";
				}
				return requestedBuild;
			}
		}
	};
	controller.fetch();
	return controller;
}

function makeBranchController(project, field, customBranchProcessor, snapshotField, snapshotSelectionChangeListener){
	var snapshot_controller = makeSnapshotController(null, snapshotField, null, null);
	renderBranchList(field, null, -1, project!=null, null);
	
	//in case if we just had to clean up data...
	if ((project == null))
		return null;
	
	var defaultBranch = -1;
	var currentBranch = null;
	var controller = {
			snapshot_controller: snapshot_controller,
			branchList: [],
			gotBranches: false,
			fetch: function(){
				controller.fetchBranchList(controller.processFetched);
			},
			processFetched:function(data){
				controller.branchList = data.branches;
				defaultBranch = getDefaultBranch(controller.branchList);
				if (defaultBranch>=0)
					currentBranch = controller.branchList[defaultBranch].name;
				controller.repaintList();
			},
			
			repaintList: function(){
				renderBranchList(field, controller.branchList, false, controller.switchBranch);
			},

			switchBranch: function(newBranch){
				currentBranch = newBranch;
				
				controller.snapshot_controller = makeSnapshotController(project, snapshotField, currentBranch, snapshotSelectionChangeListener);
//				if (branchSelectionChangeListener!=null){
//					//I can has async!
//					setTimeout(function(){
//						branchSelectionChangeListener(/*controller.getSelectedSnapshotID()*/)
//					}, 1);
//				}
			},
			
			getBranch: function(){
				return currentBranch;
			},
			
			getSnapshot: function(){
				return controller.snapshot_controller.getSelectedSnapshotID();
			},
			
			fetchBranchList: function(processBranchList) {
				$.ajax( {
					url:'/history/api?act=branches'+getProjectParameter(project),
					dataType: 'json',
					type: 'GET',
					cache: false,
					success: function (data, textStatus) {
						if (data.status != "OK") {
							setError("Error:"+data.error);
						} else {
							processBranchList(data);
						}
					},
					error: ajaxError,
					complete: function(){
						controller.gotBranches = true;
					}
				});
			}
	}
	controller.fetch();
	return controller;
}
function getDefaultBranch(data){
	var defaultBranch = -1;
	if (data.length>0){
		defaultBranch = 0;
		for (var i = 0; i < data.length; i++) {
			var branch = data[i];
			if (branch.name=='master')
				defaultBranch = i;
		}
	}
	return defaultBranch;
}

function renderBranchList(field, data, isLoading, listener){
	if (field == null)
		return;
	
	var list = $("<select size=1 style='width:100'/>");
	field.html("");
	field.append(list);
	if (data == null){
		if (!isLoading)
			list.append($("<option>----</option>"));
		else
			list.append($("<option>Loading...</option>"));
		list.prop('disabled', true);
	} else {
		list.unbind();
		list.empty();
		if (data.length==0){
			var html="<option value=''>No branches</option>";
			list.append($(html));
			list.prop('disabled', true);
		} else {
			for (var i = 0; i < data.length; i++) {
				var branch = data[i];
				list.append($("<option value=\""+branch.name+"\">"+branch.name+"</option>"));
			}
			list.prop('selectedIndex', getDefaultBranch(data));
			if (listener!=null){
				list.change(function() {
					setTimeout(function(){
						listener(list.val());
					}, 1);
				});
				list.change();
			}
			list.prop('disabled', false);
		}
	}
}

/**
 * Creates some kind of a multi-project controller for the LFA.
 * @param controllers
 */
function wrapControllers(projects, field, controllerAcceptor, switchHandler){
	renderBranchList(field, null, true, null);
	var allBranches=[];
	var projectsByBranch={};
	
	makeController = function(){
		//all Branch lst shouldd be already filled at the time of the call
		var defaultBranchID = getDefaultBranch(allBranches);
		var currentBranch = ((defaultBranchID<0)?null:allBranches[defaultBranchID]);
		var controller = {
				getActiveProjects: function(){
					var branch = controller.getCurrentBranch();
					if (branch==null)
						return [];
					var result = projectsByBranch[branch];
					if (result==null)
						return [];
					return result;
				},
				
				getCurrentBranch: function(){
					return currentBranch;//TODO
				},
				
				setBranchLFA: function(branch){
					if (branch==currentBranch)
						return;
		    		drop(SVACE_CLEANUP_SNAPSHOT);
					currentBranch = branch;
					var projects = controller.getActiveProjects()
					var branch = controller.getCurrentBranch();
					for (var i = 0; i<projects.length; i++){
						projects[i].ui.controller.switchBranch(branch);
					}
					var wait = function(){
						for (var i = 0; i<projects.length; i++){
							if (!projects[i].ui.controller.snapshot_controller.gotBuilds){
								setTimeout(wait, 10);
								myout("Fetching snapshot lists for given branch for all projects, please wait..");
								return;
							}
						}
						
						//removing projects having this branch empty from a list of valid objects...
						var toRemove = [];
						for (var i = 0; i<projects.length; i++){
							if (projects[i].ui.controller.getSnapshot()==null){
								toRemove.push(projects[i]);
							}
						}
						for (var i = 0; i<toRemove.length; i++){
							//this will also modify th version generated by getActiveProjects
							projects.splice(projects.indexOf(toRemove[i]), 1);
						}
						
						if (switchHandler!=null){
							setTimeout(switchHandler, 1);
						}
					}
					setTimeout(wait, 10);
				}
		};
		
		renderBranchList(field, allBranches, false, function(data){
			controller.setBranchLFA(data);
		});
		if (controllerAcceptor!=null){
			controllerAcceptor(controller);
		}
	}
	
	/**
	 * We should not initialize this controller until all sub-controllers have loaded their branches.
	 */
	waitForAllControllersInitialized = function (){
		for (var i = 0; i< projects.length; i++){
			var controller = projects[i].ui.controller;
			if (!controller.gotBranches){
				setTimeout(waitForAllControllersInitialized, 10);
				return;
			}
		}

		for (var i = 0; i< projects.length; i++){
			var controller = projects[i].ui.controller;
			for (var k = 0; k< controller.branchList.length; k++ ){
				var branch = controller.branchList[k];
				if (projectsByBranch[branch.name]==undefined) {
					allBranches.push(branch);
					projectsByBranch[branch.name]=[];
				}
				projectsByBranch[branch.name].push(projects[i]);
			}
		}
		//now 'allBranches' is a list of all possible branches in all projects
		makeController(allBranches);
	}
	setTimeout(waitForAllControllersInitialized, 1);
}